from datetime import datetime
from typing import Any

from colorama import Back, Fore, Style
from ..structs import Assistant, Example, Result, User

right = "❯❯❯"
left = "❮❮❮"


class NoColor:
    def __getattr__(self, *args):
        return ""


NoColor = NoColor()


def str_to_num(num_str):
    alpha = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    chars = f"0123456789{alpha}{alpha.lower()}_"
    num = 0

    for i, c in enumerate(reversed(num_str)):
        num += chars.index(c) * (len(chars) ** i)

    return num


def debug(
    obj: dict,
    color: bool = True,
    line_length: int = 120,
    whitelist: list = [str, User, Result, Assistant],
) -> None:
    B = Back if color else NoColor
    F = Fore if color else NoColor
    S = Style if color else NoColor

    type_: str = obj["type"]
    payload: Any = obj["payload"]
    timestamp: datetime = obj["timestamp"]
    app: str = obj["app"]

    time: str = timestamp.strftime("%H:%M:%S")
    message_class = payload.__class__
    message_type: str = message_class.__name__
    direction: str = left if type_ == "input" else right

    if whitelist and message_class not in whitelist:
        return

    app_colors: list[str] = [
        F.YELLOW,
        F.MAGENTA,
        F.CYAN,
        F.GREEN,
        F.BLUE,
        F.RED,
        F.LIGHTRED_EX,
        F.LIGHTGREEN_EX,
        F.LIGHTBLUE_EX,
        F.LIGHTMAGENTA_EX,
    ]

    app_color: str = app_colors[str_to_num(app) % len(app_colors)]
    direction_color: str = F.BLUE if type_ == "input" else F.GREEN
    message_type_bg: str = B.BLUE if type_ == "input" else B.GREEN

    if hasattr(payload, "content"):
        message = payload.content

        if isinstance(message, dict):
            message = message.get("response", str(message))

    elif isinstance(payload, Example):
        message = (
            f"example(user):{payload.user}\nexample(assistant):{payload.assistant}"
        )
    else:
        message: str = str(payload)

    formatted = S.RESET_ALL.join(
        [
            S.DIM + F.LIGHTWHITE_EX + f"({time})",
            S.BRIGHT + direction_color + f" {direction} ",
            S.BRIGHT + message_type_bg + F.WHITE + f" {message_type} ",
            app_color + f" {message}",
        ]
    )

    suffix = B.BLACK + S.BRIGHT + app_color + f"[{app}]" + S.RESET_ALL

    # Add app name as right-aligned suffix
    formatted = formatted + " " * max(1, line_length - len(formatted) + 20) + suffix

    # Separator if multiline
    if "\n" in formatted:
        formatted += "\n" + "―" * line_length + "\n"

    print(formatted)



from functools import wraps
import inspect
from typing import (
    Awaitable,
    Callable,
    Protocol,
    TypeVar,
)


from .structs import (
    Generate,
    User,
)

from .turbo import turbo

from .utils.args import ensure_args


__all__ = [
    "completion",
]

T = TypeVar("T")


class CompletionFn(Protocol):
    def __call__(self, *args, **kwargs) -> Awaitable[None]:
        ...


class Completion(Protocol[T]):
    def __call__(self, *args, **kwargs) -> Awaitable[T]:
        ...


ParseFn = Callable[[str], T]


# Decorator
def completion(
    parse: ParseFn[T] = lambda s: str(s),
    **opts,
) -> Callable[[CompletionFn], Completion[T]]:
    """Parameterized decorator for creating a simple generate function"""

    def wrapper(fn: CompletionFn):
        @turbo(original_fn=fn, **opts)
        async def generate(**kwargs):
            yield User(template=inspect.getdoc(fn), variables=kwargs)
            yield Generate()

        @wraps(fn)
        async def wrapped(*args, **kwargs) -> str:
            # Get cache and memory args
            cache_args = kwargs.pop("cache_args", {})
            memory_args = kwargs.pop("memory_args", {})

            # Bind args
            params = inspect.signature(fn).bind(*args, **kwargs)
            opts = params.arguments

            # Ensure args
            assert ensure_args(fn, opts)

            # Get result
            result = await generate(
                **opts,
                cache_args=cache_args,
                memory_args=memory_args,
            ).run()

            return parse(result.content)

        return wrapped

    return wrapper



from pprint import pformat
from typing import Any

from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import PythonLexer


def pprint_color(obj: Any) -> Any:
    """Pretty-print in color."""

    print(highlight(pformat(obj), PythonLexer(), Terminal256Formatter()), end="")
    return obj



import asyncio
import json
from typing import Any

import redis.asyncio as redis

from ..types.cache import BaseCache

__all__ = [
    "RedisCache",
]


class RedisCache(BaseCache):
    """Redis cache"""

    client: redis.Redis
    prefix: str
    ttl: int = 0

    async def setup(
        self,
        client: redis.Redis,
        prefix: str = "turbo_cache",
    ):
        self.client = client
        self.prefix = prefix

    def serialize(self, obj: Any) -> str:
        return json.dumps(obj)

    def deserialize(self, hashed: str) -> Any:
        return json.loads(hashed)

    def to_key(self, key) -> str:
        return f"{self.prefix}:{super().to_key(key)}"

    async def has(self, key) -> bool:
        result: int = await self.client.exists(self.to_key(key))
        return bool(result)

    async def set(self, key, value) -> None:
        _key = self.to_key(key)
        _value = self.serialize(value)

        # Set ttl
        ttl = self.ttl
        opts = {}
        if ttl:
            opts["px"] = ttl
            opts["keepttl"] = True

        await self.client.set(_key, _value, **opts)

    async def get(self, key) -> Any:
        _key = self.to_key(key)
        result = await self.client.get(_key)

        assert result, "key not found"

        return self.deserialize(result)

    async def clear(self) -> None:
        keys = [key async for key in self.client.scan_iter(f"{self.prefix}:*")]
        await asyncio.gather(*map(self.client.delete, keys))
